/*
 * Svjatoslav Commons - shared library of common functionality. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.commons.file;

import java.io.*;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.nio.file.Files.isSymbolicLink;

public class IOHelper {

    /**
     * Deletes files and directories recursively. Does not follow symlinks.
     *
     * @param file directory to delete with entire contents.
     * @throws IOException if filesystem error happens
     */
    public static void deleteRecursively(final File file) throws IOException {

        if (file.isDirectory()) {
            deleteDirectory(file);
            return;
        }

        if (file.isFile()){
            if (!file.delete()) throw new IOException("Failed to delete file: " + file);
        } else {
            if (isSymbolicLink(file.toPath()) && !file.delete())
                throw new IOException("Failed to delete symlink: " + file);
        }
    }

    private static void deleteDirectory(File file) throws IOException {
        // if file is symlink that points to directory, no not touch content
        if (!isSymbolicLink(file.toPath())){
            File[] files = file.listFiles();
            if (files == null)
                throw new IOException("Failed to read directory content for: "
                        + file);

            for (final File subFile : files)
                deleteRecursively(subFile);
        }

        if (!file.delete())
            throw new IOException("Failed to delete directory: " + file);
    }

    public static byte[] getFileContents(final File file)
            throws IOException {

        final byte[] result = new byte[(int) file.length()];
        try (final FileInputStream fileInputStream = new FileInputStream(file)) {
            if (fileInputStream.read(result) != result.length)
                throw new RuntimeException("Could not read file content:" + file);
        }
        return result;
    }

    /**
     * Expects file content to be in UTF-8 encoding.
     * @param file file to read
     * @return File content
     * @throws IOException when file reading fails.
     */
    public static String getFileContentsAsString(final File file)
            throws IOException {
        try {
            return new String(getFileContents(file), UTF_8);
        } catch (final UnsupportedEncodingException exception) {
            throw new RuntimeException(exception);
        }
    }

    /**
     * Compares new file content with old file content. If content in equal,
     * then leaves file as-is. If content differs, then overrides file with the
     * new content.
     *
     * @param file       file to potentially overwrite
     * @param newContent new content
     * @return <code>true</code> if file was overwritten.
     * @throws FileNotFoundException if file is not found.
     * @throws IOException           if error happens during file IO.
     */
    public static boolean overwriteFileIfContentDiffers(final File file,
                                                        final byte[] newContent) throws IOException {

        checkForEquality:
        {
            if (file.length() == newContent.length) {

                final byte[] oldContent = getFileContents(file);

                for (int i = 0; i < newContent.length; i++)
                    if (newContent[i] != oldContent[i])
                        break checkForEquality;

                // new file content in identical to old content
                return false;
            }
        }

        // New content differs from existing. Overwrite file.
        saveToFile(file, newContent);
        return true;
    }

    public static void saveToFile(final File file, final byte[] content)
            throws IOException {
        try (final FileOutputStream fos = new FileOutputStream(file)) {
            fos.write(content);
        }
    }

    public static void saveToFile(final File file, final String content)
            throws IOException {
        saveToFile(file, content.getBytes(UTF_8));
    }

}
