/*
 * Svjatoslav Commons - shared library of common functionality. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.commons.string;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class String2 {

    private final List<Character> chars = new ArrayList<>();

    public String2 clear(){
        chars.clear();
        return this;
    }

    public String2(String value) {
        append(value);
    }

    public String2() {
    }

    public String2 repeat(int count){
        if (count > 0) {
            String value = toString();

            for (int i = 1; i < count; i++)
                append(value);
        } else {
            clear();
        }
        return this;
    }

    public String2 prepend(final String prefix) {
        if (prefix == null)
            return this;

        int i = 0;
        for (final char c : prefix.toCharArray())
            chars.add(i++, c);

        return this;
    }

    public String2 append(final String suffix) {
        if (suffix == null)
            return this;

        for (final char c : suffix.toCharArray())
            chars.add(c);

        return this;
    }

    public String2 appendWithSeparator(String separator, final String suffix) {
        if (!isEmpty())
            append(separator);

        append(suffix);

        return this;
    }

    public String2 append(String s, int times) {
        for (int i = 0; i < times; i++) append(s);
        return this;
    }


    /**
     * Cut given amount of characters from the left of the string.
     *
     * @param cutAmount of characters to cut
     * @return part that was cut.
     */
    public String2 trimPrefix(final int cutAmount) {

        int actualCutAmount = cutAmount;

        if (actualCutAmount > getLength())
            actualCutAmount = getLength();

        if (actualCutAmount > 0) chars.subList(0, actualCutAmount).clear();

        return this;
    }

    public String2 trimPrefixIfExists(String prefix) {
        if (prefix == null)
            return this;

        if (hasPrefix(prefix))
            trimPrefix(prefix.length());

        return this;
    }

    public String2 trimSuffixIfExists(String suffix) {
        if (hasSuffix(suffix))
            trimSuffix(suffix.length());

        return this;
    }

    public String2 trimSuffix(int charsToTrim) {

        if (charsToTrim > chars.size()) {
            chars.clear();
            return this;
        }

        for (int i = 0; i < charsToTrim; i++)
            chars.remove(chars.size() - 1);

        return this;
    }

    public boolean hasSuffix(String suffix) {
        return contains(suffix, getLength() - suffix.length());
    }

    public boolean hasPrefix(String prefix) {
        return contains(prefix, 0);
    }

    public boolean contains(String fragment, int index) {
        if (index + fragment.length() > chars.size())
            return false;

        for (int i = 0; i < fragment.length(); i++)
            if (chars.get(index + i) != fragment.charAt(i))
                return false;

        return true;
    }

    public String2 enforceLength(final int targetLength) {
        if (getLength() > targetLength)
            chars.subList(targetLength, getLength()).clear();
        else if (getLength() < targetLength) {
            final int charactersToAdd = targetLength - getLength();
            for (int i = 0; i < charactersToAdd; i++)
                chars.add(' ');
        }

        return this;
    }

    public int getLength() {
        return chars.size();
    }

    public String getSubString(final int startInclusive, final int endExclusive) {
        final char[] charArray = new char[endExclusive - startInclusive];

        int j = 0;
        for (int i = startInclusive; i < endExclusive; i++) {
            charArray[j] = chars.get(i);
            j++;
        }
        return new String(charArray);
    }

    public boolean isEmpty() {
        return chars.size() == 0;
    }

    @Override
    public String toString() {
        return getSubString(0, chars.size());
    }


    public static String[] getGroups(String s, String regexp){
        Pattern pattern = Pattern.compile(regexp);
        Matcher matcher = pattern.matcher(s);

        matcher.find();
        String[] result = new String[matcher.groupCount()];

        for (int i = 0; i< result.length; i++){
            result[i] = matcher.group(i+1);
        }

        return result;
    }
}
